from copy import deepcopy
from functools import lru_cache
from typing import Callable, List, Optional, Union

from .config import ModuleConfig
from .curves import CurveConfig
from .node import Node
from .team import Team
from .tree import Tree


class Module(object):
    CONST_ROUND: int = 8
    CONST_TEAMS = "teams"
    CONST_CURVE_A = "curve_assignments"

    def __init__(self, period_input):
        self.__check_input(period_input)

        # initialze module configurations & settings
        self.config = ModuleConfig(period_input)
        self.curves = CurveConfig(period_input)

        # create decisions, sum and average i.e. the market
        self.decisions: dict = deepcopy(self.teams[0].decisions)
        self.sum: Tree = deepcopy(self.teams[0].t0)
        self.average: Tree = deepcopy(self.teams[0].t0)

        # compute market and decision average
        self.compute_market()
        self.compute_decision_avg()

    def __check_input(self, period_input):
        if isinstance(period_input, dict):
            if self.CONST_TEAMS not in period_input.keys():
                raise AttributeError(
                    "period_input <class 'dict'> has not attribute '{0}'".format(
                        self.CONST_TEAMS
                    )
                )
            if self.CONST_CURVE_A not in period_input.keys():
                raise AttributeError(
                    "period_input <class 'dict'> has not attribute '{0}'".format(
                        self.CONST_CURVE_A
                    )
                )

            # extract curve assignments & teams from json
            self.from_dict(period_input)
        else:
            raise TypeError(
                "Expected: period_input <class 'dict'>, Received: {0}".format(
                    type(period_input)
                )
            )

    def from_dict(self, period_input: dict):
        # create a Team per team input
        self.teams = [Team(team) for team in period_input[self.CONST_TEAMS]]

    def to_dict(self) -> dict:
        return {"teams": [team.to_dict() for team in self.teams]}

    def compute_market(self):
        """compute the average of all leaf nodes across the teams
        """
        # get the leaf_names from a team to ensure that
        # if they have changed -> changes are reflected in market too
        # or at least an error is raised
        leaf_names = self.teams[0].t0.leaf_names

        # compute the average of all leaf nodes across the teams
        # use leaf names to find relevant nodes
        for name in leaf_names:
            # average the values of all teams for each leaf node (name)
            scores = [team.node_t0(name).value for team in self.teams]
            sum_ = sum(scores)

            # assign the value to the market leaf
            self.sum.node(name).value = sum_
            self.average.node(name).value = sum_ / len(scores)

    def compute_decision_avg(self):
        """compute the average of all decisions across the teams
        """
        # get the key names from a team to ensure that
        # if they have changed -> changes are reflected in market too
        decision_keys = self.teams[0].decisions.keys()

        # compute the average of all decisions across the teams
        for decision in decision_keys:
            scores = [team.decisions[decision] for team in self.teams]
            average = sum(scores) / len(scores)

            self.decisions[decision] = average

    def node_t0(self, id_: int, name: str) -> Node:
        """returns the node with name: {name}
        from subtree t0 of team with id: {id_}
        """
        return self.teams[id_].node_t0(name)

    def node_t1(self, id_: int, name: str) -> Node:
        """returns the node with name: {name}
        from subtree t1 of team with id: {id_}
        """
        return self.teams[id_].node_t1(name)

    @lru_cache(maxsize=None)
    def get_curve(self, name: str) -> Callable:
        return self.curves.curves[name]

    @lru_cache(maxsize=None)
    def get_config(self, leaf: str, effect: str, value: str) -> Union[float, str, None]:
        try:
            # try to get the value of the EffectConfig
            return self.config.value(leaf, effect, value)
        except KeyError:
            # if it doesn't exist -> return None instead of raising a KeyError
            # the keys: 'MAX', 'MIN', 'WEIGHT' are allowed to not exist
            if value in ["MAX", "MIN", "WEIGHT"]:
                return None
            else:
                raise

    def _sum_product_leaf_decision(
        self, leafs: List[str], decisions: List[str], team_id: int
    ) -> float:
        """computes the sum of the multiplication of two arrays.
        multiplies the leaf value with the associated decision value
        and returns the sum.

        inputs:
        - leafs: List[str] - names of tree leafs
        - decisions: List[str] - names of decisions which influence leafs
        """
        result: float = 0.0

        for leaf, decision in zip(leafs, decisions):
            leaf_value = self.node_t0(team_id, leaf).value
            decision_value = self.teams[team_id].decisions[decision]

            value = leaf_value * decision_value
            result += round(value, self.CONST_ROUND)

        return result

    def _sum_product_leaf_config(
        self, leafs: List[str], effect: str, config: str, team_id: int
    ) -> float:
        result: float = 0.0

        for leaf in leafs:
            weight = float(self.config.value(leaf, effect, config))
            value = self.node_t1(team_id, leaf).value
            result += weight * value
        return result

    def min_max_weight(self, leaf: str, effect: str, input_: float) -> float:
        # try to get the min_, max_, weight config values
        max_: Optional[float] = self.get_config(leaf, effect, "MAX")  # type: ignore
        min_: Optional[float] = self.get_config(leaf, effect, "MIN")  # type: ignore
        weight: Optional[float] = self.get_config(leaf, effect, "WEIGHT")  # type: ignore # noqa: E501

        # if they don't exist -> None is returned
        # use default value instead of None
        if max_ is None:
            max_ = float("inf")
        if min_ is None:
            min_ = float("-inf")
        if weight is None:
            weight = 1.0

        # ensure the input_ is within bounds and weight factor adjusted
        return min(max(input_, min_), max_) * weight

    def _curve(self, leaf: str, effect: str, input_: float) -> float:
        # get the correct sensitivity curve
        curve_name = self.get_config(leaf, effect, "CURVE")
        curve = self.get_curve(curve_name)

        # compute the value on the sensitivity curve
        return curve(input_)

    def _value_change(
        self, decision: str, leaf: str, effect: str, input_: float,
    ) -> float:
        # compute the value on the sensitivity curve
        value = self._curve(leaf, effect, input_)

        # ensure the input_ is within bounds and weight factor adjusted
        return self.min_max_weight(leaf, effect, value)

    def _delta_change(
        self, decision: str, leaf: str, effect: str, team_id: int
    ) -> float:
        # compute the delta of the team's decisions compared to the market average
        team_decision = self.teams[team_id].decisions[decision]
        average = self.decisions[decision]
        delta = round(team_decision - average, self.CONST_ROUND)

        # compute the value on the sensitivity curve
        value = self._curve(leaf, effect, delta)

        # ensure the result is within bounds and weight factor adjusted
        return self.min_max_weight(leaf, effect, value)



from typing import List


class UpdateValueError(Exception):
    """
    Exception raised when user attempts to update the value
    of a node which is not a leaf node.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(
        self,
        expression,
        message="It's forbidden to update the value of non-leaf node type Nodes.",
    ):
        self.expression = expression
        self.message = message


class Node(object):
    CONST_EXCLUDED_KEYS = ["parent"]

    def __init__(
        self,
        key: str,
        name: str,
        description: str = "",
        value: float = 0,
        parent=None,
        children=None,
    ):
        self.key = key
        self.name = name
        self.description = description

        self.children = children
        self.parent = parent
        self.value = value

        # compute the base value of the node
        self.__update_value()

        # check that the keys of the children are correct
        self.__update_children_key()

    def __str__(self):
        return """Key: {0},
         Name: {1}, Description: {2},
         Value: {3}, # of Children: {4}""".format(
            self.key, self.name, self.description, self.value, len(self.children)
        )

    def to_dict(self) -> dict:
        response = {
            "key": self.key,
            "name": self.name,
            "description": self.description,
            "value": self.value,
        }
        response["children"] = [child.to_dict() for child in self.children]
        return response

    @property
    def children(self) -> List["Node"]:
        return self.__children

    @children.setter
    def children(self, children: List["Node"]):
        if children is None:
            self.__children: List["Node"] = []
        else:
            self.__children = children

        # ensure all children have the correct parent set
        for child in self.__children:
            child.parent = self

        # compute the value (since the node is no longer a leaf node)
        self.__update_value()

        # ensure the children's index is correct
        self.__update_children_key()

    @property
    def value(self) -> float:
        return self.__value

    @value.setter
    def value(self, value: float):
        # this prevents updating parent values directly
        if len(self.children) != 0 and value != 0:
            raise UpdateValueError(
                "Node: {0} has children: {1}".format(self.name, self.children),
            )

        # accept float and int else throw a TypeError
        if type(value) == float:
            self.__value = value
        elif type(value) == int:
            self.__value = float(value)
        else:
            raise TypeError(
                "Expected: value <class 'float | int'>, Received: {0}".format(
                    type(value)
                )
            )

        if hasattr(self, "parent") and self.parent:
            # trigger the recomputing all parent values
            self.parent.__update_value()

    def __update_value(self):
        # only compute value of non-leaf type Nodes
        if len(self.children) != 0:
            # this avoids the setter which would raise an exception
            # since the value of a non-leaf node is update directly
            # -> not sure if this is the best way to do it or
            # if we should just allow updating the value of non-leaf nodes
            # don't replace __value with value!
            self.__value = sum(child.value for child in self.children)

        # only update value of parent if object has a parent attribute
        if hasattr(self, "parent") and self.parent:
            self.parent.__update_value()

    def __update_children_key(self):
        # update the children keys according to their order in the array
        for index, child in enumerate(self.children):
            child.key = self.key + str(index)



import json
import os


def get_path_testfile(filename: str) -> str:
    # absolute path to the directory of the file
    abspath_filedir = os.path.dirname(os.path.abspath(__file__))

    # folder where test files are stored
    folder_testfiles = "../../files"
    folder_testfiles = os.path.join(abspath_filedir, folder_testfiles)

    # compose filepath from dir, folder_testfiles and the filename
    return os.path.abspath("{0}/{1}".format(folder_testfiles, filename))


def get_testfile_json(filename: str) -> dict:
    # setup
    filepath = get_path_testfile(filename)

    # check that file exists
    if not os.path.exists(filepath) or not os.path.isfile(filepath):
        raise FileNotFoundError("File: {0} does not exist!".format(filepath))

    # read file and assign to global variable
    with open(filepath, "r") as file:
        # read file and load json string as dict
        return json.loads(file.read())



from typing import Dict, List, Union


class EffectConfig(object):
    def __init__(self, period_input):
        self.name: str
        self.values: Dict[str, Union[str, float]] = {}
        self.__check_input(period_input)

    def __check_input(self, period_input):
        if isinstance(period_input, dict):
            self.from_dict(period_input)
        else:
            raise TypeError(
                "Expected: period_input <class 'dict'>, Received: {0}".format(
                    type(period_input)
                )
            )

    def from_dict(self, period_input: dict):
        self.name = period_input["name"]

        for param in period_input["values"]:
            key_ = param["key"]

            if key_ in ["MIN", "MAX", "WEIGHT"]:
                self.values[key_] = float(param["value"])
            else:
                self.values[key_] = param["value"]

    def value(self, key: str) -> Union[str, float]:
        return self.values[key]


class NodeConfig(object):
    def __init__(self, period_input):
        self.effects: List[EffectConfig] = []
        self.name: str
        self.__check_input(period_input)

    def __check_input(self, period_input):
        if isinstance(period_input, dict):
            self.from_dict(period_input)
        else:
            raise TypeError(
                "Expected: period_input <class 'dict'>, Received: {0}".format(
                    type(period_input)
                )
            )

    def from_dict(self, period_input: dict):
        """create a EffectConfig for each effect in the input
        An EffectConfig is composed of:
        - name: str
        - values: dict - key-value pairs containing the configuration values
        """
        self.name = period_input["name"]
        self.effects = [EffectConfig(effect) for effect in period_input["effects"]]

    def effect(self, name: str) -> EffectConfig:
        if not isinstance(name, str):
            raise TypeError(
                "Expected: name <class 'str'>, Received: {0}".format(type(name))
            )

        # retrieve the effect by its name
        effect = list(filter(lambda effect: effect.name == name, self.effects))

        # check if an entry was found, else throw an exception
        if len(effect) == 1:
            return effect[0]
        else:
            raise ValueError("No effect found for: name '{0}'".format(name))

    def value(self, effect: str, value: str) -> Union[str, float]:
        return self.effect(effect).value(value)


class ModuleConfig(object):
    def __init__(self, period_input):
        self.configurations: List[NodeConfig] = []

        # process period_input and convert into objects
        self.__check_input(period_input)

    def __check_input(self, period_input):
        if isinstance(period_input, dict):
            self.from_dict(period_input)
        else:
            raise TypeError(
                "Expected: period_input <class 'dict'>, Received: {0}".format(
                    type(period_input)
                )
            )

    def node_names(self) -> List[str]:
        return list(map(lambda node: node.name, self.configurations))

    def node(self, name: str) -> NodeConfig:
        if not isinstance(name, str):
            raise TypeError(
                "Expected: name <class 'str'>, Received: {0}".format(type(name))
            )

        # retrieve the node by its name
        node = list(filter(lambda node: node.name == name, self.configurations))

        # check if an entry was found, else throw an exception
        if len(node) == 1:
            return node[0]
        else:
            raise ValueError("No node found for: name '{0}'".format(name))

    def effect(self, name: str, effect: str) -> EffectConfig:
        return self.node(name).effect(effect)

    def value(self, name: str, effect: str, value: str) -> Union[str, float]:
        return self.effect(name, effect).value(value)

    def from_dict(self, period_input: dict):
        """create a NodeConfig object for each (Leaf-)Node in the input json.
        A NodeConfig is composed of:
        - name: str
        - effects: List[Effects] - effects which are used during the state computation
        """
        self.configurations = [NodeConfig(node) for node in period_input["config"]]



from functools import cached_property
from typing import Dict, List

from .node import Node


class Tree(object):
    CONST_CACHED_PROPERTIES = ["leaf_names"]

    def __init__(self, root=None):
        self.name_key_map: Dict[str, str] = {}
        self.key_node_map: Dict[str, Node] = {}
        self.root: Node = root

    def to_dict(self) -> dict:
        response = self.root.to_dict()
        response["children"] = [child.to_dict() for child in self.root.children]
        return response

    @property
    def root(self):
        return self.__root

    @root.setter
    def root(self, tree_input):
        if isinstance(tree_input, Node):
            # don't use -> only used for testing
            self.__root = tree_input
        elif isinstance(tree_input, dict):
            self.__root = self.from_dict(tree_input)
        else:
            raise TypeError(
                "Expected: tree_input <class 'Node | dict'>, Received: {0}".format(
                    type(tree_input)
                )
            )

    @cached_property
    def leaf_names(self) -> List[str]:
        leaf_names = []

        for node in self.key_node_map.values():
            if len(node.children) == 0:
                leaf_names.append(node.name)

        return leaf_names

    def node(self, name: str) -> Node:
        """Retruns a Node with matching name if it exists in the Tree.

        Parameters:
        - name: str - the name of the Node

        Returns:
        - node: Node - the Node in the Tree with a matching name

        Raises:
        - KeyError: if name does not exist
        """
        key = self.node_key(name)
        return self.__node_by_key(key)

    def node_key(self, name: str) -> str:
        return self.name_key_map[name]

    def __node_by_key(self, key: str) -> Node:
        return self.key_node_map[key]

    def __invalidate_cache(self):
        """Invalidates all cache entries of the class by
        clearing the lru_cache for the function calls with arguments
        and by clearing all cached_properties.
        """
        # clear all cached properties
        for prop in self.CONST_CACHED_PROPERTIES:
            self.__invalidate_cached_property(prop)

    def __invalidate_cached_property(self, name: str):
        """Deletes the dict entry of a cached property.
        """
        if name in self.__dict__:
            del self.__dict__[name]

    def from_dict(self, tree_input: dict) -> Node:
        """Creates a Tree from a dict of nodes.

        Parameters:
        - root: dict - the input JSON as a dictionary
        - LOG: bool - flag to decide if the creation of the nodes shall be logged

        Returns:
        - root: Node - the root Node of the tree including all children
        """
        # invalidate cache entries (since Tree changes)
        self.__invalidate_cache()

        # create the Tree recursively from the JSON input
        return self.__rec_create_nodes_from_dict("0", tree_input["tree_input"])

    def __rec_create_nodes_from_dict(self, index: str, tree_input: dict) -> Node:
        # base case: node has no children -> it is a leaf
        if len(tree_input["children"]) == 0:
            # create a leaf node and assign the value
            leaf = Node(
                key=index,
                name=tree_input["name"],
                description=tree_input["description"],
                value=tree_input["value"],
            )

            # create mapping name -> key
            self.name_key_map[tree_input["name"]] = index

            # create mapping key -> node
            self.key_node_map[index] = leaf
            return leaf

        # recursion: node has children -> it is a parent
        else:
            children = []
            parent_index = index

            # create parent without children
            parent = Node(
                key=parent_index,
                name=tree_input["name"],
                description=tree_input["description"],
            )

            # create mapping name -> key
            self.name_key_map[tree_input["name"]] = index

            for index_, child_json in enumerate(tree_input["children"]):
                child_index = parent_index + str(index_)

                # recursively call each child and see if it has children
                child = self.__rec_create_nodes_from_dict(child_index, child_json)
                children.append(child)

            parent.children = children

            # create mapping key -> node
            self.key_node_map[parent.key] = parent
            return parent



from typing import Callable


class CurveConfig(object):
    CONST_CURVES = "curve_assignments"
    CONST_SLOT_KEY = "curve_slot_key"
    CONST_CURVE = "curve_assigned"
    CONST_COEFFICIENTS = "coefficients"

    def __init__(self, period_input):
        # initialize curves
        self.curves: dict = {}

        # process period_input and convert into objects
        self.__check_input(period_input)

    def __check_input(self, period_input):
        if isinstance(period_input, dict):
            if self.CONST_CURVES not in period_input.keys():
                raise AttributeError(
                    "period_input <class 'dict'> has not attribute '{0}'".format(
                        self.CONST_CURVES
                    )
                )

            for curve in period_input[self.CONST_CURVES]:
                keys = curve.keys()

                if self.CONST_SLOT_KEY not in keys:
                    raise AttributeError(
                        "curve_input <class 'dict has not attribute '{0}'".format(  # noqa: E501
                            self.CONST_SLOT_KEY
                        )
                    )

                if self.CONST_CURVE not in keys:
                    raise AttributeError(
                        "curve_input <class 'dict has not attribute '{0}'".format(  # noqa: E501
                            self.CONST_CURVE
                        )
                    )

                curve_assigned_keys = curve[self.CONST_CURVE].keys()
                if self.CONST_COEFFICIENTS not in curve_assigned_keys:
                    raise AttributeError(
                        "curve_assigned_input <class 'dict has not attribute '{0}'".format(  # noqa: E501
                            self.CONST_COEFFICIENTS
                        )
                    )

            self.from_dict(period_input)

    def from_dict(self, period_input: dict):
        for curve in period_input[self.CONST_CURVES]:
            # the name of the curve in the Python Model i.e. the curve slot key
            curve_name = curve[self.CONST_SLOT_KEY]

            # the function coefficients as array (e.g., [1, 0, 2])
            coefficients = curve[self.CONST_CURVE][self.CONST_COEFFICIENTS]

            # the create lambda function
            self.curves[curve_name] = self.coefficients_to_lambda_function(coefficients)

    def coefficients_to_lambda_function(self, coefficients) -> Callable:
        """Create a lambda function given a curve configuration.

        Parameters:
        - coefficients: an array of coefficients (e.g., [2,3,4])

        Returns: a lambda function representing the curve

        Example: lambda x: 2 + 3*x + 4*x**2
        """
        return lambda x: sum(
            coefficient * x ** index for index, coefficient in enumerate(coefficients)
        )



from copy import deepcopy

from .node import Node
from .tree import Tree


class Team(object):
    CONST_ID = "team_id"
    CONST_DEC = "decisions"
    CONST_TREE = "tree_input"

    def __init__(self, period_input):
        # initialize empty decisions dict
        self.decisions: dict = {}
        self.__check_input(period_input)

    def __check_input(self, period_input):
        if isinstance(period_input, dict):
            if self.CONST_ID not in period_input.keys():
                raise AttributeError(
                    "period_input <class 'dict'> has not attribute '{0}'".format(
                        self.CONST_ID
                    )
                )

            if self.CONST_DEC not in period_input.keys():
                raise AttributeError(
                    "period_input <class 'dict'> has not attribute '{0}'".format(
                        self.CONST_DEC
                    )
                )

            if self.CONST_TREE not in period_input.keys():
                raise AttributeError(
                    "period_input <class 'dict'> has not attribute '{0}'".format(
                        self.CONST_TREE
                    )
                )

            # convert dict to object
            self.from_dict(period_input)
        else:
            raise TypeError(
                "Expected: period_input <class 'dict'>, Received: {0}".format(
                    type(period_input)
                )
            )

    def from_dict(self, period_input: dict):
        # assign id
        self.id = int(period_input[self.CONST_ID])

        # transform decisions input into dict
        for decision in period_input[self.CONST_DEC]:
            self.decisions[decision["key"]] = decision["value"]

        # create tree from tree_input (current state t0)
        self.t0 = Tree(period_input)

        # create a copy of the tree t0 (next state t1)
        self.t1 = deepcopy(self.t0)

    def to_dict(self) -> dict:
        return {"team_id": self.id, "tree": self.t1.root.to_dict()}

    def node_t0(self, name: str) -> Node:
        return self.t0.node(name)

    def node_t1(self, name: str) -> Node:
        return self.t1.node(name)



from setuptools import find_packages, setup

setup(
    name="bms-tree-modules",
    version="0.1.1",
    packages=find_packages(exclude=["tests", "files", "src/utils"]),
    package_data={"": ["LICENSE", "NOTICES"]},
    python_requires=">=3.8",
    setup_requires=[],
    install_requires=["typing"],
    extras_require={
        "dev": ["pylama", "black", "flake8"],
        "tests": ["pytest", "pytest-watch"],
    },
    include_package_data=True,
    author="Moritz Eck",
    author_email="moritz.eck@gmail.com",
    description="The package has been created as part of the BMS20 project at the Institue of Banking & Finance and Institute of Informatics (University of Zurich). It allows to create a tree data structure recursively from a given input json/dict. The package delivers the game building blocks for the various components: Node, Tree, Team, Config, Curves, Module.",  # noqa: E501
    url="https://gitlab.uzh.ch/uzh-bf/sim/banking-game-2020",
    project_urls={
        "Documentation": "https://gitlab.uzh.ch/uzh-bf/sim/banking-game-2020",
        "Source Code": "https://gitlab.uzh.ch/uzh-bf/sim/banking-game-2020",
    },
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
    ],
)
